!*********************************************************************************************************************************************************
!> Module: OpticalProperties
!>
!>
!>  This module contains the subroutines for calculating the epsilon tensor using the generalized Drude-Lorentz model
!>
!>
!>
!>  The following subroutines and functions are included in this module:
!>
!>      - function g:                               calculate g function defined by A.B. Kuzmenko, Rev. of Scien. Instr. 76, 083108 (2005), (eq. 13)
!>      - subroutine vde:                           calculate the real and imaginary part of the epsilon tensor elements
!>      - subroutine reflectance:                   calculate the reflectance within the ac-plane
!>      - subroutine CalcEigewnvectors:             calculate the eigenvectors of a sym. 2x2 complex matrix using analytic func. determined by Mathematica
!>      - subroutine DetermineRotationAngles:       calculate the rotation angles for real and imaginary part, determine the corresponding diagonal
!>                                                  matrix entries and calculate R_p using Koch formula
!>      - subroutine InverseMatrix:                 calculate the inverse of a 2x2 complex matrix
!>
!>
!> Copyright (C) 2009 - 2014
!>
!> I. Physikalisches Institut, Universitaet zu Koeln
!>
!> Produced for the CATS project and MnWO4 Paper
!>
!>
!> fortran module containing the subroutine for "generalized Drude-Lorentz model"
!>
!> Who           When        What
!>
!> T. Moeller    07/07/2009  Initial version
!>
!*********************************************************************************************************************************************************
Module OpticalProperties

    implicit none
    integer :: NumberXValues                                                                !< number exp. data points
    integer :: number_osc                                                                   !< number of osc. for the generalized Drude-Lorentz model
    integer :: number_PolAngle                                                              !< number of polarization angles
    integer :: freq_counter                                                                 !< counter for frequencies
    integer :: switchflag                                                                   !< working variable: used for consistency
    real*8 :: pi                                                                            !< pi
    real*8 :: eps_inf_xx                                                                    !< epsilon_\infty for epsilon_xx matrix element
    real*8 :: eps_inf_xz                                                                    !< epsilon_\infty for epsilon_xz matrix element
    real*8 :: eps_inf_zz                                                                    !< epsilon_\infty for epsilon_zz matrix element
    real*8 :: LastReRotAngle, LastImRotAngle                                                !< rotation angles for the previous frequencies
    real*8, allocatable, dimension(:) :: omega                                              !< frequency points
    real*8, allocatable, dimension(:) :: PolarizationAngles                                 !< polarization angles
    real*8, allocatable, dimension(:) :: w0, wp, Gamma, theta                               !< eigenfrequencies, plasma frequencies, damping, and angle
                                                                                            !< for the generalized Drude-Lorentz model
    real*8, allocatable, dimension(:) :: sqrt_func                                          !< working variable: used for consistency
    real*8 :: w2, w12, gamma2                                                               !< parameter for frequency dependent damping
    real*8, allocatable, dimension(:, :) :: Aij                                              !< oscillator strengths for var-dielectric function


    contains


        !>************************************************************************************************************************************************
        !> subroutine: g
        !>
        !> calculate g function defined by A.B. Kuzmenko, Rev. of Scien. Instr. 76, 083108 (2005), (eq. 13)
        !>
        !>
        !> input variables:     x:                  i
        !>
        !>                      y:                  i
        !>
        !> output variables:    g:                  calculated value for g
        !>
        !>
        !> \author Thomas Moeller
        !>
        !> \date 07.07.2009
        !>
        function g(x, y)

            implicit none
            real*8 :: x, y                                                                  !<
            real*8 :: g                                                                     !<

            if (dabs(x - y) < 1.d-12) then
                g = (x + y) * dlog(dabs(x + y))
            else
                g = (x + y) * dlog(dabs(x + y)) + (x - y) * dlog(dabs(x - y))
            endif
            return
        end function g


        !>************************************************************************************************************************************************
        !> subroutine: vde
        !>
        !> calculate the real and imaginary part of the epsilon tensor elements 
        !>
        !>
        !> input variables:     w_index:            index for frequency
        !>
        !> output variables:    epsvarxx:           variational contribution to the eps_xx element 
        !>
        !>                      epsvarxz:           variational contribution to the eps_xz element 
        !>
        !>                      epsvarzz:           variational contribution to the eps_zz element 
        !>
        !>
        !> \author Thomas Moeller
        !>
        !> \date 07.07.2009
        !>
        subroutine vde(w_index, epsvarxx, epsvarxz, epsvarzz)

            implicit none
            integer :: i                                                                    !< loop variable
            integer :: w_index                                                              !< index of frequency
            real*8 :: w, w_i, w_im1, w_ip1                                                  !< parameters for frequency points
            real*8 :: eps1, eps2                                                            !< real and imaginary part of epsilon function
            complex(8) :: epsvarxx, epsvarxz, epsvarzz                                      !< variational contribution to the epsilon tensor elements


            epsvarxx = dcmplx(0.d0, 0.d0)
            epsvarxz = dcmplx(0.d0, 0.d0)
            epsvarzz = dcmplx(0.d0, 0.d0)
            Do i = 2, (NumberXValues - 1)
                w = omega(w_index)
                w_i = omega(i)
                w_im1 = omega(i - 1)
                w_ip1 = omega(i + 1)


                !< define epsilon2
                if (w_im1 < w .and. w <= w_i) then
                    eps2 = (w - w_im1) / (w_i - w_im1)
                elseif (w_i < w .and. w < w_ip1) then
                    eps2 = (w_ip1 - w) / (w_ip1 - w_i)
                else
                    eps2 = 0.d0
                endif


                !< define epsilon1
                eps1 = (1.d0/pi) * ( (g(w, w_im1)) / (w_i - w_im1) - ((w_ip1 - w_im1) * g(w, w_i)) / ((w_i - w_im1) * (w_ip1 - w_i)) &
                                   + (g(w, w_ip1)) / (w_ip1 - w_i))


                !< construct final epsilon
                epsvarxx = epsvarxx + Aij(1, i) * dcmplx(eps1, eps2)
                epsvarxz = epsvarxz + Aij(2, i) * dcmplx(eps1, eps2)
                epsvarzz = epsvarzz + Aij(3, i) * dcmplx(eps1, eps2)
            end Do
            return
        end subroutine vde


        !>************************************************************************************************************************************************
        !> subroutine: reflectance
        !>
        !> calculate the reflectance within the ac-plane
        !>
        !>
        !> input variables:     w_index:            index for frequency
        !>
        !>                      angle_index:        index for polarization angle
        !>
        !>                      w:                  frequency
        !>
        !>                      angle:              polarization angle
        !>
        !> output variables:    value:              calculated reflectance
        !>
        !>
        !> \author Thomas Moeller
        !>
        !> \date 07.07.2009
        !>
        subroutine reflectance(w_index, angle_index, w, angle, value)

            implicit none
            integer :: w_index                                                              !< index of frequency
            integer :: angle_index                                                          !< index of polarization angle
            integer :: i, k                                                                 !< loop variables
            real*8 :: w                                                                     !< frequency
            real*8 :: value                                                                 !< calculated reflectance
            real*8 :: angle                                                                 !< polarization angle
            real*8 :: theta_rad                                                             !< theta is in radians
            real*8 :: w1                                                                    !< working variable: eigenfrequency
            real*8 :: gamma1                                                                !< working variable: damping
            real*8, dimension(2, 2) :: Unit_matrix                                          !< 2x2 unit matrix
            complex(8) :: eps11, eps12, eps21, eps22                                        !< working variables: components of the epsilon tensor
            complex(8) :: DrudeLorentz                                                      !< working variables: argument of the sum for one oscillator
            complex(8) :: A1, A2                                                            !< working variables
            complex(8) :: epsvar11, epsvar12, epsvar22                                      !< working variables: contributions of variational ansatz
            complex(8), dimension(2) :: dir_vector, help_vector                             !< working variables: used for calculation of reflectance
            complex(8), dimension(2) :: ev1, ev2                                            !< eigenvectors
            complex(8), dimension(2, 2) :: epsilon_tensor                                   !< epsilon tensor
            complex(8), dimension(2, 2) :: Inverse_epsilon_tensor                           !< inverse of epsilon tensor
            complex(8), dimension(2, 2) :: Eigenvectors_epsilon_tensor                      !< eigenvector matrix
            complex(8), dimension(2, 2) :: diagonal_matrix                                  !< diagonal (rotated) form of the epsilon tensor
            complex(8), dimension(2, 2) :: sqrt_diagonal_matrix                             !< square root of diagonal form
            complex(8), dimension(2, 2) :: new_diagonal_matrix                              !< working variable: new diagonal form of epsilon tensor
            complex(8), dimension(2, 2) :: upper_refl_matrix                                !< working variable: first matrix for Fresnel formula
            complex(8), dimension(2, 2) :: lower_refl_matrix                                !< working variable: second matrix for Fresnel formula
            complex(8), dimension(2, 2) :: Inverse_lower_refl_matrix                        !< working variable: inverse of second matrix


            !< reset output variable
            value = 0.d0


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< calculate matrix elements of the epsilon tensor
            eps11 = eps_inf_xx
            eps12 = eps_inf_xz
            eps22 = eps_inf_zz
            k = 4
            Do i = 1, number_osc
                if (Gamma(i) >= 0.d0) then                                                  !< loop over all oscillators
                    DrudeLorentz = ((wp(i)**2) / (w0(i)**2 - w**2 - dcmplx(0.d0, 1.d0) * w * Gamma(i)))
                else


                    !< apply correct derived ansatz eq. 18 (Grueninger) with additional parameters: w12, w2, gamma2
                    ! w1 = w0(i)
                    ! gamma1 = dabs(Gamma(i))
                    ! A1 = (w1**2 + w12**2 - w**2 - dcmplx(0.d0, 1.d0) * w * gamma1)
                    ! A2 = (w2**2 + w12**2 - w**2 - dcmplx(0.d0, 1.d0) * w * gamma2)
                    ! DrudeLorentz = ((wp(i)**2) / (A1 - (w12**4) / A2))


                    !< apply correct derived ansatz eq. 18 (Grueninger) with additional parameters: w12, w2, gamma2
                    w1 = w0(i)
                    gamma1 = dabs(Gamma(i))
                    A1 = (w1**2 - w**2 - dcmplx(0.d0, 1.d0) * w * gamma1)
                    A2 = (w2**2 - w**2 - dcmplx(0.d0, 1.d0) * w * gamma2)
                    DrudeLorentz = ((wp(i)**2) / (A1 - (w12**4) / A2))
                endif
                theta_rad = theta(i) * (pi/180.d0)                                          !< convert degree to rad
                eps11 = eps11 + DrudeLorentz * (dcos(theta_rad)**2)
                eps12 = eps12 + DrudeLorentz * (dsin(theta_rad) * dcos(theta_rad))
                eps22 = eps22 + DrudeLorentz * (dsin(theta_rad)**2)
                k = k + 4
            end Do


            !< get contribution var. dielectric function
            call vde(w_index, epsvar11, epsvar12, epsvar22)


            !< add varational contribution to epsilon functions
            eps11 = eps11 + epsvar11
            eps12 = eps12 + epsvar12
            eps22 = eps22 + epsvar22
            eps21 = eps12


            !< check, if Im[eps_11] and Im[eps_22] < 0
            if (dimag(eps11) < 0.d0) then
                eps11 = dcmplx(dreal(eps11), 0.d0)
            endif
            if (dimag(eps22) < 0.d0) then
                eps22 = dcmplx(dreal(eps22), 0.d0)
            endif


            !< write epsilon tensor to file
            if (angle_index == 1) then
                write(80,*) w, dreal(eps11)
                write(81,*) w, dimag(eps11)
                write(82,*) w, dreal(eps12)
                write(83,*) w, dimag(eps12)
                write(84,*) w, dreal(eps22)
                write(85,*) w, dimag(eps22)
            endif


            !< define epsilon tensor
            epsilon_tensor = dcmplx(0.d0, 0.d0)
            epsilon_tensor(1, 1) = eps11
            epsilon_tensor(1, 2) = eps12
            epsilon_tensor(2, 1) = eps21
            epsilon_tensor(2, 2) = eps22

            ! Debug:
            !   print*,'frequency = ',w
            !   print*,epsilon_tensor(1, 1), epsilon_tensor(1, 2)
            !   print*,epsilon_tensor(2, 1), epsilon_tensor(2, 2)


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< determine eigenvectors of epsilon tensor (expression taken from Mathematica)
            !< determine eigenvectors of epsilon tensor (expression taken from Mathematica)
            call CalcEigewnvectors(epsilon_tensor, ev1, ev2)


            !< construct eigenvector matrix
            Eigenvectors_epsilon_tensor = dcmplx(0.d0, 0.d0)
            Eigenvectors_epsilon_tensor(1, 1) = ev2(1)
            Eigenvectors_epsilon_tensor(2, 1) = ev2(2)
            Eigenvectors_epsilon_tensor(1, 2) = ev1(1)
            Eigenvectors_epsilon_tensor(2, 2) = ev1(2)

            ! Debug:
        	! print*,"Eigenvectors_epsilon_tensor(1,:) = ", Eigenvectors_epsilon_tensor(1, 1), Eigenvectors_epsilon_tensor(1, 2)
        	! print*,"Eigenvectors_epsilon_tensor(2,:) = ", Eigenvectors_epsilon_tensor(2, 1), Eigenvectors_epsilon_tensor(2, 2)


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< determine inverse of eigenvector matrix
            Inverse_epsilon_tensor = dcmplx(0.d0, 0.d0)
            call InverseMatrix(Eigenvectors_epsilon_tensor, Inverse_epsilon_tensor)

            ! Debug:
            !	print*,Inverse_epsilon_tensor(1, 1), Inverse_epsilon_tensor(1, 2)
            !	print*,Inverse_epsilon_tensor(2, 1), Inverse_epsilon_tensor(2, 2)


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< bring epsilon tensor on diagonal form
            diagonal_matrix = dcmplx(0.d0, 0.d0)
            diagonal_matrix = matmul(matmul(Inverse_epsilon_tensor, epsilon_tensor), Eigenvectors_epsilon_tensor)

            ! Debug:
            !	print*,diagonal_matrix(1, 1), diagonal_matrix(1, 2)
            !	print*,diagonal_matrix(2, 1), diagonal_matrix(2, 2)


            !< determine rotation angle
            if (angle_index == 1) then
                call DetermineRotationAngles(w, epsilon_tensor)
            endif


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< take square root of diagonal elements
            sqrt_diagonal_matrix = dcmplx(0.d0, 0.d0)
            sqrt_diagonal_matrix = diagonal_matrix
            sqrt_diagonal_matrix(1, 1) = cdsqrt(sqrt_diagonal_matrix(1, 1))
            sqrt_diagonal_matrix(2, 2) = cdsqrt(sqrt_diagonal_matrix(2, 2))

            ! Debug:
            !	print*,sqrt_diagonal_matrix(1, 1), sqrt_diagonal_matrix(1, 2)
            !	print*,sqrt_diagonal_matrix(2, 1), sqrt_diagonal_matrix(2, 2)


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< rotate matrix back
            new_diagonal_matrix = dcmplx(0.d0, 0.d0)
            new_diagonal_matrix = matmul(matmul(Eigenvectors_epsilon_tensor, sqrt_diagonal_matrix), Inverse_epsilon_tensor)

            ! Debug:
            !	print*,new_diagonal_matrix(1, 1), new_diagonal_matrix(1, 2)
            !	print*,new_diagonal_matrix(2, 1), new_diagonal_matrix(2, 2)


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< determine upper and lower part of the reflectance matrix


            !< define unity matrix
            Unit_matrix = 0.d0
            Unit_matrix(1, 1) = 1.d0
            Unit_matrix(2, 2) = 1.d0


            !< define upper matrix
            upper_refl_matrix = dcmplx(0.d0, 0.d0)
            upper_refl_matrix = Unit_matrix - new_diagonal_matrix


            !< define lower matrix
            lower_refl_matrix = dcmplx(0.d0, 0.d0)
            lower_refl_matrix = Unit_matrix + new_diagonal_matrix


            !< determine inverse of lower_refl_matrix
            Inverse_lower_refl_matrix = dcmplx(0.d0, 0.d0)
            call InverseMatrix(lower_refl_matrix, Inverse_lower_refl_matrix)

            ! Debug:
            !	print*,upper_refl_matrix(1,1),upper_refl_matrix(1,2)
            !	print*,upper_refl_matrix(2,1),upper_refl_matrix(2,2)

            !	print*,lower_refl_matrix(1,1),lower_refl_matrix(1,2)
            !	print*,lower_refl_matrix(2,1),lower_refl_matrix(2,2)

            !	print*,Inverse_lower_refl_matrix(1,1),Inverse_lower_refl_matrix(1,2)
            !	print*,Inverse_lower_refl_matrix(2,1),Inverse_lower_refl_matrix(2,2)


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< calulate reflectance
            dir_vector = (/dcmplx(dcos(angle * (pi/180.d0)),0.d0),dcmplx(dsin(angle * (pi/180.d0)),0.d0)/)
            help_vector = matmul(matmul(upper_refl_matrix, Inverse_lower_refl_matrix), dir_vector)
            value = dreal(dot_product(help_vector, help_vector))
            return
        end subroutine reflectance


        !>************************************************************************************************************************************************
        !> subroutine: CalcEigewnvectors
        !>
        !> calculate the eigenvectors of a symmetric 2x2 complex matrix using analytic functions determined by Mathematica
        !>
        !>
        !> input variables:     matrix:             matrix which has to be diagonalized
        !>
        !> output variables:    ev1:                first eigenvector
        !>
        !>                      ev2:                second eigenvector
        !>
        !>
        !> \author Thomas Moeller
        !>
        !> \date 07.07.2009
        !>
        subroutine CalcEigewnvectors(matrix, ev1, ev2)

            implicit none
            complex(8) :: eps11, eps12, eps22                                               !< working variables: matrix elements
            complex(8), dimension(2) :: ev1, ev2                                            !< eigenvectors 1 and 2
            complex(8), dimension(2, 2) :: matrix                                           !< symmetric matrix


            !< get component of matirx
            eps11 = matrix(1, 1)
            eps12 = matrix(1, 2)
            eps22 = matrix(2, 2)


            !< EV1: Mathematica formula for non-normalized eigenvector 1 and a separat normalization
            ev1(1) = -((-eps11 + eps22 + cdsqrt(eps11**2 + 4 * eps12**2 - 2 * eps11 * eps22 + eps22**2))/(2*eps12))
            ev1(2) = 1.d0
            ev1(:) = 1.d0 / dsqrt(cdabs(dot_product(ev1(:), ev1(:)))) * ev1(:)


            !< EV1: Mathematica formula for already normalized eigenvector 1
            !ev1(1) = (eps11 - eps22 - cdsqrt(eps11**2 + dcmplx(4.d0, 0.d0) * eps12**2 - dcmplx(2.d0, 0.d0) * eps11 * eps22 + eps22**2)) &
            !         /(eps12 * cdsqrt(dcmplx(4.d0, 0.d0) + cdabs((-eps11 + eps22 + cdsqrt(eps11**2 + dcmplx(4.d0, 0.d0) * eps12**2 &
            !         - dcmplx(2.d0, 0.d0) * eps11 * eps22 + eps22**2))/eps12)**2))
            !ev1(2) = 2.d0 / cdsqrt(dcmplx(4.d0, 0.d0) + cdabs((-eps11 + eps22 + cdsqrt(eps11**2 + dcmplx(4.d0, 0.d0) * eps12**2 &
            !         - dcmplx(2.d0, 0.d0) * eps11 * eps22 + eps22**2)) / eps12)**2)


            !< EV2: Mathematica formula for non-normalized eigenvector 2 and a separat normalization
            ev2(1) = -((-eps11 + eps22 - cdsqrt(eps11**2 + 4 * eps12**2 - 2 * eps11 * eps22 + eps22**2))/(2*eps12))
            ev2(2) = 1.d0
            ev2(:) = 1 / dsqrt(cdabs(dot_product(ev2(:), ev2(:)))) * ev2(:)


            !< EV2: Mathematica formula for already normalized eigenvector 2
            !ev2(1) = (eps11 - eps22 + cdsqrt(eps11**2 + dcmplx(4.d0, 0.d0) * eps12**2 - dcmplx(2.d0, 0.d0) * eps11 * eps22 + eps22**2)) &
            !         /(eps12 * cdsqrt(dcmplx(4.d0, 0.d0) + cdabs((eps11 - eps22 + cdsqrt(eps11**2 + dcmplx(4.d0, 0.d0) * eps12**2 &
            !         - dcmplx(2.d0, 0.d0) * eps11 * eps22 + eps22**2))/eps12)**2))
            !ev2(2) = 2.d0 /cdsqrt(dcmplx(4.d0, 0.d0) + cdabs((eps11 - eps22 + cdsqrt(eps11**2 + dcmplx(4.d0, 0.d0) * eps12**2 &
            !         - dcmplx(2.d0, 0.d0) * eps11 * eps22 + eps22**2))/eps12)**2)
            return
        end subroutine CalcEigewnvectors


        !>************************************************************************************************************************************************
        !> subroutine: DetermineRotationAngles
        !>
        !> calculate the rotation angles for real and imaginary part, determine the corresponding diagonal matrix entries and calculate R_p using Koch
        !>
        !>
        !> input variables:     w:                  frequency
        !>
        !>                      angle:              polarization angle
        !>
        !> output variables:    None
        !>
        !>
        !> \author Thomas Moeller
        !>
        !> \date 07.07.2009
        !>
        subroutine DetermineRotationAngles(w, matrix)

            implicit none
            real*8 :: w                                                                     !< frequency
            real*8 :: ReRotAngle                                                            !< value of current rotation angle of the real part
            real*8 :: ImRotAngle                                                            !< value of current rotation angle of the imaginary part
            real*8 :: h1                                                                    !< working variable for eigenvector check
            real*8, dimension(2, 2) :: S, Smin1                                             !< rotation matrices for tilting
            real*8, dimension(2, 2) :: test                                                 !< working variable for unit check of inverse matrix
            complex(8) :: ww                                                                !< working variable: used for sign of square root
            complex(8), dimension(2) :: ev1, ev2                                            !< eigenvectors for real and imaginary part
            complex(8), dimension(2, 2) :: ReMatrix, ImMatrix                               !< real and imaginary part of the epsilon tensor
            complex(8), dimension(2, 2) :: matrix                                           !< copy of epsilon tensor
            complex(8), dimension(2, 2) :: eigenvectormatrix, Inverse_eigenvectormatrix     !< eigenvector matrix and inverse of eigenvector matrix
            complex(8), dimension(2, 2) :: diagonal_matrix                                  !< diagonal (rotated) matrix for real and imaginary part
            complex(8), dimension(2, 2) :: Epsrot                                           !< used for calculation of R_p


            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< real part


            !< determine real part of the epsilon tensor
            ReMatrix = dcmplx(0.d0, 0.d0)
            ReMatrix(1, 1) = dcmplx(dreal(matrix(1, 1)), 0.d0)
            ReMatrix(1, 2) = dcmplx(dreal(matrix(1, 2)), 0.d0)
            ReMatrix(2, 1) = dcmplx(dreal(matrix(2, 1)), 0.d0)
            ReMatrix(2, 2) = dcmplx(dreal(matrix(2, 2)), 0.d0)


            !< determine eigenvectors
            ev1 = dcmplx(0.d0, 0.d0)
            ev2 = dcmplx(0.d0, 0.d0)
            call CalcEigewnvectors(ReMatrix, ev1, ev2)

            ! Debug:
            ! print*,"EV(Re[epsion_tensor])(1,1) = ", dreal(ev1(1))
            ! print*,"EV(Re[epsion_tensor])(1,2) = ", dreal(ev1(2))
            ! print*,"EV(Re[epsion_tensor])(2,1) = ", dreal(ev2(1))
            ! print*,"EV(Re[epsion_tensor])(2,2) = ", dreal(ev2(2))


            !< check orthonormality of eigenvectors
            h1 = dreal(dot_product(ev1, ev1))
            if (dabs(h1 - 1.d0) > 1.d-10) then
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  The first calculated eigenvector of Re[epsilon] is not normalized to 1'
                print*,' '
                print*,' '
                print*,'  ev1 = ', ev1
                print*,'  ev1.ev1 = ', h1
                stop
            endif
            h1 = dreal(dot_product(ev2, ev2))
            if (dabs(h1 - 1.d0) > 1.d-10) then
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  The second calculated eigenvector of Re[epsilon] is not normalized to 1'
                print*,' '
                print*,' '
                print*,'  ev2 = ', ev2
                print*,'  ev2.ev2 = ', h1
                stop
            endif
            h1 = dreal(dot_product(ev1, ev2))
            if (dabs(h1) > 1.d-10) then
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  The two calculated eigenvector of Re[epsilon] are not orthogonal to each other!'
                print*,' '
                print*,' '
                print*,'  ev1 = ', ev1
                print*,'  ev2 = ', ev2
                print*,'  ev1.ev2 = ', h1
                stop
            endif


            !< check for consistency
            ww = cdsqrt(matrix(1, 1)**2 + 4 * matrix(1, 2)**2 - 2 * matrix(1, 1) * matrix(2, 2) + matrix(2, 2)**2)
            sqrt_func(freq_counter) = dimag(ww)
            if (sqrt_func(freq_counter - 1) > 0.d0 .and. sqrt_func(freq_counter) < 0.d0 &
                .or. sqrt_func(freq_counter - 1) < 0.d0 .and. sqrt_func(freq_counter) > 0.d0) then
                if (dabs(sqrt_func(freq_counter - 1) - sqrt_func(freq_counter)) > 1.d1) then
                    switchflag = 1 - switchflag

                    ! Debug:
                    ! print*,'w_jump = ', w
                endif
            endif

            ! Debug:
            ! write(12345,*) w, dreal(ww), dimag(ww)


            !< deterimine rotation angle for real part of the epsilon tensor
            if (switchflag == 1) then
                ReRotAngle = datan2(dreal(ev2(2)), dreal(ev2(1)))
            else
                ReRotAngle = datan2(dreal(ev1(2)), dreal(ev1(1)))
            endif


            !< check, if rotation angle function runs out of interval [0,pi)
            if (LastReRotAngle /= -1.d99) then

                ! Debug:
                ! print*,"dabs(ReRotAngle - LastReRotAngle) = ", dabs(ReRotAngle - LastReRotAngle)


                if (dabs(ReRotAngle - LastReRotAngle) >= (pi * 0.92d0)) then
                    ReRotAngle = ReRotAngle + pi
                endif
            endif


            !< write rotation angle to file
            write(3010, *) w, (ReRotAngle - pi) * 180.d0/pi, ReRotAngle * 180.d0/pi, (ReRotAngle + pi) * 180.d0/pi


            !< construct eigenvectormatrix
            if (switchflag == 1) then
                eigenvectormatrix(1, 1) = ev2(1)
                eigenvectormatrix(1, 2) = ev2(2)
                eigenvectormatrix(2, 1) = ev1(1)
                eigenvectormatrix(2, 2) = ev1(2)
            else
                eigenvectormatrix(1, 1) = ev1(1)
                eigenvectormatrix(1, 2) = ev1(2)
                eigenvectormatrix(2, 1) = ev2(1)
                eigenvectormatrix(2, 2) = ev2(2)
            endif


            !< calculate inverse of eigenvector matrix
            Inverse_eigenvectormatrix = dcmplx(0.d0, 0.d0)
            call InverseMatrix(eigenvectormatrix, Inverse_eigenvectormatrix)


            !< rotate real part of epsion tensor to diagonal form
            diagonal_matrix = dcmplx(0.d0, 0.d0)
            diagonal_matrix = matmul(matmul(eigenvectormatrix, ReMatrix), Inverse_eigenvectormatrix)
            if (dabs(dreal(diagonal_matrix(1, 2))) > 1.d-10 .or. dabs(dreal(diagonal_matrix(2, 1))) > 1.d-10) then
                diagonal_matrix = matmul(matmul(Inverse_eigenvectormatrix, ReMatrix), eigenvectormatrix)
            endif


            !< check off-diagonal elements
            if (dabs(dreal(diagonal_matrix(1, 2))) > 1.d-10 .or. dabs(dreal(diagonal_matrix(2, 1))) > 1.d-10) then
                test = dreal(matmul(Inverse_eigenvectormatrix, eigenvectormatrix))
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  Off-diagonal elements of rotated Re[epsilon] tensor are not zero!'
                print*,' '
                print*,' '
                print*,'  w = ', w
                print*,'  ev1 = ', ev1
                print*,'  ev2 = ', ev2
                print*,' '
                print*,'  eigenvectormatrix(1, :) = ', dreal(eigenvectormatrix(1, :))
                print*,'  eigenvectormatrix(2, :) = ', dreal(eigenvectormatrix(2, :))
                print*,' '
                print*,'  Inverse_eigenvectormatrix(1, :) = ', dreal(Inverse_eigenvectormatrix(1, :))
                print*,'  Inverse_eigenvectormatrix(2, :) = ', dreal(Inverse_eigenvectormatrix(2, :))
                print*,' '
                print*,'  unity(1, :) = ', test(1, :)
                print*,'  unity(2, :) = ', test(2, :)
                print*,' '
                print*,'  diagonal_matrix(1, :) = ', dreal(diagonal_matrix(1, :))
                print*,'  diagonal_matrix(2, :) = ', dreal(diagonal_matrix(2, :))
                print*,' '
                stop
            endif


            !< write diagonal elements to file
            write(3001, *) w, dreal(diagonal_matrix(1, 1))
            write(3003, *) w, dreal(diagonal_matrix(2, 2))



            !---------------------------------------------------------------------------------------------------------------------------------------------
            !< imaginary part


            !< determine real part of the epsilon tensor
            ImMatrix = dcmplx(0.d0, 0.d0)
            ImMatrix(1, 1) = dcmplx(dimag(matrix(1, 1)), 0.d0)
            ImMatrix(1, 2) = dcmplx(dimag(matrix(1, 2)), 0.d0)
            ImMatrix(2, 1) = dcmplx(dimag(matrix(2, 1)), 0.d0)
            ImMatrix(2, 2) = dcmplx(dimag(matrix(2, 2)), 0.d0)


            !< determine eigenvectors
            ev1 = dcmplx(0.d0, 0.d0)
            ev2 = dcmplx(0.d0, 0.d0)
            call CalcEigewnvectors(ImMatrix, ev1, ev2)

            ! Debug:
            ! print*,"EV(Im[epsion_tensor])(1,1) = ", dreal(ev1(1))
            ! print*,"EV(Im[epsion_tensor])(1,2) = ", dreal(ev1(2))
            ! print*,"EV(Im[epsion_tensor])(2,1) = ", dreal(ev2(1))
            ! print*,"EV(Im[epsion_tensor])(2,2) = ", dreal(ev2(2))


            !< check orthonormality of eigenvectors
            h1 = dreal(dot_product(ev1, ev1))
            if (dabs(h1 - 1.d0) > 1.d-10) then
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  The first calculated eigenvector of Im[epsilon] is not normalized to 1'
                print*,' '
                print*,' '
                print*,'  ev1 = ', ev1
                print*,'  ev1.ev1 = ', h1
                stop
            endif
            h1 = dreal(dot_product(ev2, ev2))
            if (dabs(h1 - 1.d0) > 1.d-10) then
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  The second calculated eigenvector of Im[epsilon] is not normalized to 1'
                print*,' '
                print*,' '
                print*,'  ev2 = ', ev2
                print*,'  ev2.ev2 = ', h1
                stop
            endif
            h1 = dreal(dot_product(ev1, ev2))
            if (dabs(h1) > 1.d-10) then
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  The two calculated eigenvector of Im[epsilon] are not orthogonal to each other!'
                print*,' '
                print*,' '
                print*,'  ev1 = ', ev1
                print*,'  ev2 = ', ev2
                print*,'  ev1.ev2 = ', h1
                stop
            endif


            !< deterimine rotation angle for real part of the epsilon tensor
            ImRotAngle = datan2(dreal(ev2(2)), dreal(ev2(1)))
            if (LastImRotAngle /= -1.d99) then
                if (dabs(ImRotAngle - LastImRotAngle) >= (pi * 0.82d0)) then
                    ImRotAngle = ImRotAngle - pi
                endif
            endif
            !switchflag = 2
            !if (switchflag == 1) then
            !    ImRotAngle = datan2(dreal(ev2(2)), dreal(ev2(1)))
            !else
            !    ImRotAngle = datan2(dreal(ev1(2)), dreal(ev1(1)))
            !endif


            !< write rotation angle to file
            write(3011, *) w, (ImRotAngle - pi) * 180.d0/pi, ImRotAngle * 180.d0/pi, (ImRotAngle + pi) * 180.d0/pi


            !< construct eigenvectormatrix
            eigenvectormatrix(1, 1) = ev1(1)
            eigenvectormatrix(1, 2) = ev1(2)
            eigenvectormatrix(2, 1) = ev2(1)
            eigenvectormatrix(2, 2) = ev2(2)


            !< calculate inverse of eigenvector matrix
            Inverse_eigenvectormatrix = dcmplx(0.d0, 0.d0)
            call InverseMatrix(eigenvectormatrix, Inverse_eigenvectormatrix)


            !< rotate real part of epsion tensor to diagonal form
            diagonal_matrix = dcmplx(0.d0, 0.d0)
            diagonal_matrix = matmul(matmul(eigenvectormatrix, ImMatrix), Inverse_eigenvectormatrix)
            if (dabs(dreal(diagonal_matrix(1, 2))) > 1.d-10 .or. dabs(dreal(diagonal_matrix(2, 1))) > 1.d-10) then
                diagonal_matrix = matmul(matmul(Inverse_eigenvectormatrix, ImMatrix), eigenvectormatrix)
            endif


            !< check off-diagonal elements
            if (dabs(dreal(diagonal_matrix(1, 2))) > 1.d-10 .or. dabs(dreal(diagonal_matrix(2, 1))) > 1.d-10) then
                print*,' '
                print*,' '
                print*,'Error in subroutine DetermineRotationAngles:'
                print*,'  Off-diagonal elements of rotated Im[epsilon] tensor are not zero!'
                print*,' '
                print*,' '
                print*,'  w = ', w
                print*,'  ev1 = ', ev1
                print*,'  ev2 = ', ev2
                print*,' '
                print*,'  eigenvectormatrix(1, :) = ', dreal(eigenvectormatrix(1, :))
                print*,'  eigenvectormatrix(2, :) = ', dreal(eigenvectormatrix(2, :))
                print*,' '
                print*,'  Inverse_eigenvectormatrix(1, :) = ', dreal(Inverse_eigenvectormatrix(1, :))
                print*,'  Inverse_eigenvectormatrix(2, :) = ', dreal(Inverse_eigenvectormatrix(2, :))
                print*,' '
                print*,'  unity(1, :) = ', test(1, :)
                print*,'  unity(2, :) = ', test(2, :)
                print*,' '
                print*,'  diagonal_matrix(1, :) = ', dreal(diagonal_matrix(1, :))
                print*,'  diagonal_matrix(2, :) = ', dreal(diagonal_matrix(2, :))
                print*,' '
                stop
            endif

            ! Debug:
        	! print*,"ReRotAngle = ", ReRotAngle
        	! print*,"ImRotAngle = ", ImRotAngle


            !< write diagonal elements
            write(3002, *) w, dreal(diagonal_matrix(1, 1))
            write(3004, *) w, dreal(diagonal_matrix(2, 2))


            !< save rotation angles
            LastReRotAngle = ReRotAngle
            LastImRotAngle = ImRotAngle


            !< tilt epsilon tensor around tilting angle
            Epsrot = matmul(matmul(S, matrix), Smin1)
            return
        end subroutine DetermineRotationAngles


        !>************************************************************************************************************************************************
        !> subroutine: InverseMatrix
        !>
        !> calculate the inverse of a 2x2 symmetric complex matrix using analytic functions determined by Mathematica
        !>
        !>
        !> input variables:     matrix:             matrix which has to inverted
        !>
        !> output variables:    Inverse_tensor:     inverted matrix
        !>
        !>
        !> \author Thomas Moeller
        !>
        !> \date 07.07.2009
        !>
        subroutine InverseMatrix(matrix, Inverse_tensor)

            implicit none
            complex(8) :: eps11, eps12, eps21, eps22                                        !< working variables: matrix elements
            complex(8), dimension(2, 2) :: matrix                                           !< input variable: matrix
            complex(8), dimension(2, 2) :: Inverse_tensor                                   !< output variable: inverse of matrix

            eps11 = matrix(1, 1)
            eps12 = matrix(1, 2)
            eps21 = matrix(2, 1)
            eps22 = matrix(2, 2)

            Inverse_tensor = dcmplx(0.d0, 0.d0)
            Inverse_tensor(1, 1) =   eps22/(-eps12 * eps21 + eps11 * eps22)
            Inverse_tensor(1, 2) = -(eps12/(-eps12 * eps21 + eps11 * eps22))
            Inverse_tensor(2, 1) = -(eps21/(-eps12 * eps21 + eps11 * eps22))
            Inverse_tensor(2, 2) =   eps11/(-eps12 * eps21 + eps11 * eps22)
            return
        end subroutine InverseMatrix
end Module OpticalProperties
!*********************************************************************************************************************************************************


!*********************************************************************************************************************************************************
!>
!>
!> Main Program
!>
!>
program DrudeLorentzGeneral
    !< calculation of the refelctivity using the generalized Drude-Lorentz model

    use OpticalProperties

    implicit none
    integer :: i, j                                                                         !< loop variables
    integer :: alloc_status, dealloc_status                                                 !< status variables for (de-)allocation
    integer :: dummy                                                                        !< dummy variable
    real*8 :: w                                                                             !< working variable: current frequency w
    real*8 :: angle_chi                                                                     !< working variable: current polarization angle
    real*8, allocatable, dimension(:) :: refl                                               !< working variable: final reflectance for frequency w


    pi = 4.d0 * datan(1.d0)                                                                 !< pi


    !< open files
    open(21,file = "in.txt")                                                                !< open parameter file
    open(22,file = "DataIn.dat")                                                            !< open experimental-point file
    open(23,file = "FitFunctionValues.dat")                                                 !< open file for fit function values


    !< read parameter from file
    read(21,*) NumberXValues                                                                !< read number of exp. data points
    read(21,*) number_PolAngle                                                              !< read number of polarization angles


    ! deallocate/allocate memory
    if (allocated(PolarizationAngles)) then
        deallocate(PolarizationAngles, Aij, refl, omega, stat = dealloc_status)
        if (dealloc_status /= 0) then
            print '(" ")'
            print '(2x,"Error in module DrudeLorentzGeneral:")'
            print '(4x,"Cannot deallocate variables omega, PolarizationAngles, refl and Aij.")'
            print '(" ")'
            stop '  Program aborted!'
        endif
    endif
    allocate(omega(NumberXValues), PolarizationAngles(number_PolAngle), Aij(number_PolAngle, NumberXValues), refl(number_PolAngle), stat = alloc_status)
    if (alloc_status /= 0) then
        print '(" ")'
        print '(2x,"Error in module DrudeLorentzGeneral:")'
        print '(4x,"Cannot allocate variables omega, PolarizationAngles, refl and Aij.")'
        print '(" ")'
        stop '  Program aborted!'
    endif
    omega = 0.d0
    PolarizationAngles = 0.d0
    Aij = 0.d0
    refl = 0.d0


    !< read in polarization angles
    Do i = 1, number_PolAngle
        read(21,*) PolarizationAngles(i)
    end Do


    !< read values for epsilon^\infty and number of oscillators
    read(21,*) eps_inf_xx                                                                   !< read epsilon_xx
    read(21,*) eps_inf_xz                                                                   !< read epsilon_xz
    read(21,*) eps_inf_zz                                                                   !< read epsilon_zz
    read(21,*) number_osc                                                                   !< read number of oscillators


    ! deallocate/allocate memory
    if (allocated(w0)) then
        deallocate(w0, wp, Gamma, theta, stat = dealloc_status)
        if (dealloc_status /= 0) then
            print '(" ")'
            print '(2x,"Error in module DrudeLorentzGeneral:")'
            print '(4x,"Cannot deallocate variable w0, wp, Gamma, theta. ")'
            print '(" ")'
            stop '  Program aborted!'
        endif
    endif
    allocate(w0(number_osc), wp(number_osc), Gamma(number_osc), theta(number_osc), sqrt_func(NumberXValues), stat = alloc_status)
    if (alloc_status /= 0) then
        print '(" ")'
        print '(2x,"Error in module DrudeLorentzGeneral:")'
        print '(4x,"Cannot allocate variable w0, wp, Gamma, theta. ")'
        print '(" ")'
        stop '  Program aborted!'
    endif
    w0 = 0.d0
    wp = 0.d0
    Gamma = 0.d0
    theta = 0.d0
    sqrt_func = 0.d0
    w2 = 0.d0
    w12 = 0.d0
    gamma2 = 0.d0


    !< read parameters for each oscillators from file
    Do i = 1, number_osc                                                                    !< loop over all oscillators
        read(21,*) w0(i), wp(i), Gamma(i), theta(i)
    end Do
    read(21,*) w2, w12, gamma2                                                              !< read in parameter for frequency dependent damping


    !< read in parameters for var-dielectric function and frequencies from file
    read(21,*) dummy
    Do j = 1, NumberXValues
        read(22,*) omega(j)                                                                 !< read frequency form file
        Do i = 1, 3
            read(21,*) Aij(i, j)
        end Do
    end Do
    close(21)                                                                               !< close and delete parameter file

    ! Debug:
    !    print*,'DrudeLorentzGeneral:'
    !    print*,'NumberXValues = ', NumberXValues
    !    print*,'number_PolAngle = ', number_PolAngle
    !    print*,'PolarizationAngles = ', PolarizationAngles
    !    print*,'eps_inf_xx = ',eps_inf_xx
    !    print*,'eps_inf_xz = ',eps_inf_xz
    !    print*,'eps_inf_zz = ',eps_inf_zz
    !    print*,'number_osc = ',number_osc
    !    print*,'w0(1:number_osc) = ',w0(1:number_osc)
    !    print*,'wp(1:number_osc) = ',wp(1:number_osc)
    !    print*,'Gamma(1:number_osc) = ',Gamma(1:number_osc)
    !    print*,'theta(1:number_osc) = ',theta(1:number_osc)
    !    print*,'w2 = ', w2
    !    print*,'w12 = ', w12
    !    print*,'gamma2 = ', gamma2
    !    print*,' '


    !< open output files for final epsilon component functions
    open(80, file = "epsilon-tensor__Re-eps11.dat")
    open(81, file = "epsilon-tensor__Im-eps11.dat")
    open(82, file = "epsilon-tensor__Re-eps12.dat")
    open(83, file = "epsilon-tensor__Im-eps12.dat")
    open(84, file = "epsilon-tensor__Re-eps22.dat")
    open(85, file = "epsilon-tensor__Im-eps22.dat")
    open(3001, file = "Diag__Re-eps_xx.dat")                                                !< open file for real part of diagonal part of eps_xx
    open(3002, file = "Diag__Im-eps_xx.dat")                                                !< open file for imaginary part of diagonal part of eps_xx
    open(3003, file = "Diag__Re-eps_zz.dat")                                                !< open file for real part of diagonal part of eps_zz
    open(3004, file = "Diag__Im-eps_zz.dat")                                                !< open file for imaginary part of diagonal part of eps_zz
    open(3010, file = "Re-rotation-angle.dat")                                              !< open file for rotation angle of real part
    open(3011, file = "Im-rotation-angle.dat")                                              !< open file for rotation angle of imaginary part
    Do i = 1, NumberXValues                                                                 !< loop over all frequency points
        refl = 0.d0
        freq_counter = i
        Do j = 1, number_PolAngle                                                           !< loop over all polarization angles
            angle_chi = PolarizationAngles(j)
            w = omega(i)


            !< call subroutine for determine reflectance
            call reflectance(i, j, w, angle_chi, refl(j))                                   !< calculate reflectance
        end Do


        !< write fit function value to file
        write(23,*) refl(:)                                                                 !< write reflection to output file
    end Do


    !< close files
    close(22)                                                                               !< close and delete experimental-point file
    close(23)                                                                               !< close output file
    close(80)
    close(81)
    close(82)
    close(83)
    close(84)
    close(85)
    close(3001)                                                                             !< close file for real part of diagonal part of eps_xx
    close(3002)                                                                             !< close file for imaginary part of diagonal part of eps_xx
    close(3003)                                                                             !< close file for real part of diagonal part of eps_zz
    close(3004)                                                                             !< close file for imaginary part of diagonal part of eps_zz
    close(3010)                                                                             !< close file for rotation angle of real part
    close(3011)                                                                             !< close file for rotation angle of imaginary part
end program DrudeLorentzGeneral
!---------------------------------------------------------------------------------------------------------------------------------------------------------

